import pickle
from abc import ABCMeta
from contextlib import AbstractContextManager
from pathlib import Path
from typing import Type, TypeVar


_Model = TypeVar('_Model', bound='AbstractModel')


class AbstractModel(AbstractContextManager, metaclass=ABCMeta):
    def __enter__(self: _Model) -> _Model:
        return self

    def __getstate__(self):
        return self.__dict__.copy()

    @classmethod
    def load(cls: Type[_Model], path: Path) -> _Model:
        if path.suffix == '.dvc':
            import dvc.api
            # problem with dvc.api.open: https://github.com/iterative/dvc/issues/4667
            model = pickle.loads(dvc.api.read(path.with_suffix(''), mode='rb'))
        else:
            with path.open('rb') as f:
                model = pickle.load(f)
        if not isinstance(model, cls):
            raise Exception(f"Model at {path} is not an instance of {cls}")
        return model

    def save(self, path: Path) -> None:
        if path.exists():
            raise Exception(f"Model saving path exists: {path}")

        with path.open('wb') as f:
            pickle.dump(self, f, pickle.HIGHEST_PROTOCOL)



from abc import ABCMeta, abstractmethod
from typing import Generic, Iterable, TypeVar

from tdm.abstract.datamodel import AbstractTreeDocumentContent
from tdm.datamodel import TalismanDocument

from .processor import AbstractDocumentProcessor

_DocumentContent = TypeVar('_DocumentContent', bound=AbstractTreeDocumentContent)
_Processor = TypeVar('_Processor', bound=AbstractDocumentProcessor)


class AbstractTrainer(Generic[_Processor, _DocumentContent], metaclass=ABCMeta):

    @abstractmethod
    def train(self,
              train_docs: Iterable[TalismanDocument[_DocumentContent]],
              dev_docs: Iterable[TalismanDocument[_DocumentContent]] = None) \
            -> _Processor:
        pass



from abc import abstractmethod
from typing import Generic, Type, TypeVar

from .model import AbstractModel

_Model = TypeVar('_Model', bound=AbstractModel)
_Wrapper = TypeVar('_Wrapper', bound='AbstractModelWrapper')


class AbstractModelWrapper(AbstractModel, Generic[_Model]):
    def __init__(self, model: _Model):
        self._model = model

    def __enter__(self):
        self._model.__enter__()
        return self

    def __exit__(self, *exc):
        self._model.__exit__(*exc)

    @classmethod
    def from_config(cls: Type[_Wrapper], model: _Model, config: dict) -> _Wrapper:
        kwargs = cls.parse_config(config)
        return cls(model, **kwargs)

    @staticmethod
    @abstractmethod
    def parse_config(config: dict) -> dict:
        pass



import json
from pathlib import Path
from typing import Union


def check_path_existence(path: Path):
    if not path.exists():
        raise Exception(f"Path {path} does not exist")


def check_path_absense(path: Path):
    if path.exists():
        raise Exception(f"Path {path} already exists")


def check_or_create_empty_path(path: Path):
    if not path.exists():
        path.mkdir(parents=True)
    else:
        try:
            next(path.rglob("*"))
            raise Exception("Working path must not exist or must be empty")
        except StopIteration:
            pass


def read_json(path: Union[str, Path]):
    with open(path, "r", encoding="utf-8") as f:
        return json.load(f)



import os
from abc import abstractmethod
from contextlib import AbstractContextManager
from typing import Dict, Optional

from sgqlc.endpoint.requests import RequestsEndpoint


class AbstractGQLClient(AbstractContextManager):
    @abstractmethod
    def execute(self, query, variables=None, operation_name=None, extra_headers=None, timeout=None):
        pass

    @staticmethod
    def _configure_gql_client(uri: str, headers: Optional[Dict] = None) -> RequestsEndpoint:
        env_timeout = os.getenv('GQL_API_TIMEOUT', None)
        timeout = float(env_timeout) if env_timeout is not None else None
        return RequestsEndpoint(uri, headers, timeout=timeout)



import logging
import os
from contextlib import AbstractContextManager
from typing import Dict, Iterable, Optional, Tuple

from requests import Timeout
from sgqlc.operation import Operation

from .gql_clients import AbstractGQLClient, KeycloakAwareGQLClient, NoAuthGQLClient
from .schemas.api_schema import ConceptPropertyValueType, ConceptType, Query

logger = logging.getLogger(__name__)


class TalismanAPIAdapter(AbstractContextManager):
    def __init__(self, gql_uri: str):
        self._gql_uri = gql_uri
        self._gql_client: Optional[AbstractGQLClient] = None

    def __enter__(self):
        self._check_api()
        if os.getenv("KEYCLOAK_AUTH_URL") is None:
            self._gql_client = NoAuthGQLClient(self._gql_uri)
        else:
            self._gql_client = KeycloakAwareGQLClient(self._gql_uri)

        logger.info(f"{type(self._gql_client)} will be used")

        self._gql_client.__enter__()
        return self

    def __exit__(self, *exc):
        self._gql_client.__exit__(*exc)
        self._gql_client = None

    def _check_api(self):
        logger.info(f"used correct version of API")
        return True

    def get_base_types(self, dictionary: bool = False, regexp: bool = False, pretrained_nercmodels: bool = False
                       ) -> Tuple[Tuple[ConceptType, ...], Tuple[ConceptPropertyValueType, ...]]:
        op = Operation(Query)

        concept_types: ConceptType = op.list_concept_type
        property_value_types: ConceptPropertyValueType = op.list_concept_property_value_type

        concept_types.id()
        property_value_types.id()

        if dictionary:
            concept_types.dictionary()
            property_value_types.dictionary()

        if regexp:
            concept_types.regexp()
            property_value_types.regexp()

        if pretrained_nercmodels:
            concept_types.pretrained_nercmodels()
            property_value_types.pretrained_nercmodels()

        ret = self.gql_call(op)
        return tuple(ret.list_concept_type), tuple(ret.list_concept_property_value_type)

    def execute_easy_query(self, name: str, params: Iterable[str], variables: Optional[Dict] = None):
        op = Operation(Query)
        if variables:
            op[name](**variables).__fields__(*params)
        else:
            op[name].__fields__(*params)
        return self.gql_call(op)[name]

    def gql_call(self, sgqlc_operation: Operation, variables: Optional[dict] = None, raise_on_timeout: bool = True):
        try:
            return sgqlc_operation + self._gql_client.execute(sgqlc_operation, variables=variables)
        except Timeout as e:
            logger.error('Timeout while query processing', exc_info=e, extra={'query': sgqlc_operation})
            if raise_on_timeout:
                raise e
        except Exception as e:
            logger.error('Some exception was occured during query processing.', exc_info=e,
                         extra={'query': sgqlc_operation})
            raise e



__all__ = [
    'AbstractCategoryAwareDocumentProcessor',
    'AbstractDocumentProcessor',
    'AbstractTrainer'
]

from .category import AbstractCategoryAwareDocumentProcessor
from .processor import AbstractDocumentProcessor
from .trainer import AbstractTrainer



from abc import ABCMeta, abstractmethod
from typing import Callable, Dict, Generic, Iterator, Optional, Type, TypeVar, Union

from tdm.abstract.datamodel import AbstractFact, AbstractTreeDocumentContent, FactStatus
from tdm.datamodel import TalismanDocument, TalismanSpan
from tdm.datamodel.fact import ConceptFact, ValueFact

EntityFactType = Union[Type[ConceptFact], Type[ValueFact]]
Label2FactType = Dict[str, EntityFactType]

_Content = TypeVar('_Content', bound=AbstractTreeDocumentContent)


class AbstractReader(Generic[_Content], metaclass=ABCMeta):
    @abstractmethod
    def read(self) -> Iterator[TalismanDocument[_Content]]:
        pass


def label_span2fact(default_type: EntityFactType,
                    label2facttype: Optional[Label2FactType] = None,
                    status: FactStatus = FactStatus.APPROVED) \
        -> Callable[[str, TalismanSpan], AbstractFact]:
    mapping = dict(label2facttype) if label2facttype is not None else {}

    def create_fact(label: str, span: TalismanSpan) -> AbstractFact:
        return mapping.get(label, default_type)(None, status, label, None, [span])

    return create_fact



__all__ = [
    'ModelTypeFactory',
    'AbstractCompositeModel', 'AbstractConfigConstructableModel', 'AbstractModelWrapper', 'AbstractMultipleConfigConstructableModel',
    'AbstractCategoryAwareDocumentProcessor', 'AbstractDocumentProcessor', 'AbstractTrainer',
    'ImmutableBaseModel',
    'AbstractUpdatableModel', 'AbstractUpdate', 'UpdatableModelWrapper', 'UpdateMode'
]

from .configuration import ModelTypeFactory
from .model import AbstractCompositeModel, AbstractConfigConstructableModel, AbstractModelWrapper, AbstractMultipleConfigConstructableModel
from .processor import AbstractCategoryAwareDocumentProcessor, AbstractDocumentProcessor, AbstractTrainer
from .schema import ImmutableBaseModel
from .updatable import AbstractUpdatableModel, AbstractUpdate, UpdatableModelWrapper, UpdateMode



from collections import defaultdict
from typing import Callable, Dict, Generic, List, Optional, Sequence, Tuple, Type, TypeVar

from tdm.abstract.datamodel import AbstractTreeDocumentContent
from tdm.datamodel import TalismanDocument

from tp_interfaces.abstract.model import AbstractCompositeModel
from tp_interfaces.abstract.schema import ImmutableBaseModel
from .processor import AbstractDocumentProcessor

_DocumentContent = TypeVar('_DocumentContent', bound=AbstractTreeDocumentContent)
_Config = TypeVar('_Config', bound=ImmutableBaseModel)
_Category = TypeVar('_Category')


class AbstractCategoryAwareDocumentProcessor(
    AbstractCompositeModel[AbstractDocumentProcessor[_Config, _DocumentContent]],
    AbstractDocumentProcessor[_Config, _DocumentContent], Generic[_Config, _DocumentContent, _Category]
):
    def __init__(self,
                 category2processor: Dict[_Category, AbstractDocumentProcessor[_Config, _DocumentContent]],
                 categorizer: Callable[[TalismanDocument[_DocumentContent]], _Category]
                 ):
        self._category2processor = dict(category2processor)
        super().__init__(self._category2processor.values())
        self._models: Tuple[AbstractDocumentProcessor[_Config, _DocumentContent], ...]
        self._categorizer = categorizer

    def process_doc(self, document: TalismanDocument[_DocumentContent], config: _Config) -> TalismanDocument[_DocumentContent]:
        return self.process_docs([document], config)[0]

    def process_docs(self, documents: Sequence[TalismanDocument[_DocumentContent]], config: _Config) \
            -> Tuple[TalismanDocument[_DocumentContent], ...]:
        category2doc = defaultdict(dict)
        processed_docs: List[Optional[TalismanDocument[_DocumentContent]]] = [None] * len(documents)

        for i, doc in enumerate(documents):
            category = self._categorizer(doc)
            category2doc[category].update({i: doc})

        for category, docs in category2doc.items():
            processor = self._category2processor[category]
            docs_ = processor.process_docs(tuple(docs.values()), config)
            for id_, doc in zip(docs, docs_):
                processed_docs[id_] = doc

        return tuple(processed_docs)

    @property
    def config_type(self) -> Type[_Config]:
        return self._models[0].config_type



from tp_interfaces.knowledge_base.interfaces import DisambiguationKB


class _Singleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(_Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class KBManager(metaclass=_Singleton):
    def __init__(self):
        self._kb = None

    @property
    def knowledge_base(self) -> DisambiguationKB:
        if self._kb is None:
            raise AttributeError("Knowledge bae is not initialized")
        return self._kb

    @knowledge_base.setter
    def knowledge_base(self, kb: DisambiguationKB):
        if self._kb is not None:
            raise AttributeError("Knowledge base already initialized")
        self._kb = kb

    @knowledge_base.deleter
    def knowledge_base(self):
        self._kb = None



from typing import Optional

from sgqlc.endpoint.requests import RequestsEndpoint

from .abstract import AbstractGQLClient


class NoAuthGQLClient(AbstractGQLClient):
    def __init__(self, gql_uri: str):
        self._gql_uri = gql_uri
        self._gql_client: Optional[RequestsEndpoint] = None

    def __enter__(self):
        self._gql_client = self._configure_gql_client(self._gql_uri)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._gql_client = None
        self._gql_client = None

    def execute(self, query, variables=None, operation_name=None, extra_headers=None, timeout=None):
        return self._gql_client(query, variables, operation_name, extra_headers, timeout)
